/*
 * Copyright (c) 2015-2018,2021 LAAS/CNRS
 * All rights reserved.
 *
 * Redistribution  and  use  in  source  and binary  forms,  with  or  without
 * modification, are permitted provided that the following conditions are met:
 *
 *   1. Redistributions of  source  code must retain the  above copyright
 *      notice and this list of conditions.
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice and  this list of  conditions in the  documentation and/or
 *      other materials provided with the distribution.
 *
 * THE SOFTWARE  IS PROVIDED "AS IS"  AND THE AUTHOR  DISCLAIMS ALL WARRANTIES
 * WITH  REGARD   TO  THIS  SOFTWARE  INCLUDING  ALL   IMPLIED  WARRANTIES  OF
 * MERCHANTABILITY AND  FITNESS.  IN NO EVENT  SHALL THE AUTHOR  BE LIABLE FOR
 * ANY  SPECIAL, DIRECT,  INDIRECT, OR  CONSEQUENTIAL DAMAGES  OR  ANY DAMAGES
 * WHATSOEVER  RESULTING FROM  LOSS OF  USE, DATA  OR PROFITS,  WHETHER  IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR  OTHER TORTIOUS ACTION, ARISING OUT OF OR
 * IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 *                                           Anthony Mallet on Wed Jun  3 2015
 */
#include <sys/time.h>
#include <sys/types.h>

#include <assert.h>
#include <errno.h>
#include <inttypes.h>
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#if defined(_WIN32) || defined(_WIN64)
# include <winsock2.h>
# define poll(f, n, t)	WSAPoll(f, n, t)
#else
# include <sys/socket.h>

# include <poll.h>
#endif

#include "libgenomix.h"


/* --- local data ---------------------------------------------------------- */

static void	http_data(int fd, struct http_rqst *rqst);
static char *	http_readline(int fd, struct http_buffer *buf);
static char *	http_read(int fd, struct http_buffer *buf, size_t len);

static ssize_t	bufread(int fd, char *buf, size_t count);
static int	bufcat(struct http_buffer *buf, const char *data, int quoted);
static int	bufappend(struct http_buffer *buf, const char *data, size_t n);
static int	bufcat_uint64_t(struct http_buffer *buf, int neg, uint64_t i);


/* --- http_get ------------------------------------------------------------ */

int
http_get(socket_s fd, struct http_rqst *rqst,
         const char *session, const char *cmd, const char *args[])
{
  const char *key, *value;

  rqst->args.start = rqst->args.end = rqst->args.data;
  bufcat(&rqst->args, cmd, 1);
  bufcat(&rqst->args, "?s=", 0);
  bufcat(&rqst->args, session, 0);
  while(args) {
    key = *args++;
    if (!key) break;
    value = *args++;
    bufcat(&rqst->args, "&", 0);
    bufcat(&rqst->args, key, 1);
    if (value) {
      bufcat(&rqst->args, "=", 0);
      bufcat(&rqst->args, value, 1);
    }
  }

  return http_transaction(fd, rqst, "get", rqst->args.data, NULL);
}


/* --- http_reply ---------------------------------------------------------- */

int
http_reply(socket_s fd, struct http_rqst *rqst)
{
  int s;

  /* process events until done */
  while(rqst->state != http_eof && rqst->state != http_ok)
    http_data(fd, rqst);

  if (rqst->state == http_ok) {
    s = rqst->status;
  } else {
    rqst->body = "broken connection";
    rqst->content_length = strlen(rqst->body);
    s = 400;
  }

  return s;
}


/* --- http_event ---------------------------------------------------------- */

int
http_event(socket_s fd, struct http_event *evt)
{
  char *l, *p;

  evt->event = evt->data = -1;
  if (evt->buffer.start > evt->buffer.data) {
    memmove(evt->buffer.data, evt->buffer.start,
            evt->buffer.end - evt->buffer.start + 1);
    evt->buffer.end -= evt->buffer.start - evt->buffer.data;
    evt->buffer.start = evt->buffer.data;
  }

  while(1) {
    l = http_readline(fd, &evt->buffer);
    if (!l) return -1;
    if (*l == '\0') return 0;

    p = strchr(l, ':');
    if (!p) continue;
    *p++ = '\0';
    p += strspn(p, " \t");

    if (!strcmp(l, "event"))
      evt->event = p - evt->buffer.data;
    else if (!strcmp(l, "data"))
      evt->data = p - evt->buffer.data;
  }

  return 0;
}


/* --- http_transaction ---------------------------------------------------- */

int
http_transaction(socket_s fd, struct http_rqst *rqst,
                 const char *method, const char *uri, const char *body)
{
  const char *p;
  size_t len;

  /* send request */
  len = body ? strlen(body) : 0;

  rqst->uri.start = rqst->uri.end = rqst->uri.data;
  bufcat(&rqst->uri, method, 0);
  bufcat(&rqst->uri, " ", 0);
  bufcat(&rqst->uri, uri, 0);
  bufcat(&rqst->uri, " HTTP/1.0\n", 0);
  if (len > 0) {
    bufcat(&rqst->uri,
           "Content-Type: application/x-www-form-urlencoded\n"
           "Content-Length: ", 0);
    bufcat_uint64_t(&rqst->uri, 0, len);
    bufcat(&rqst->uri, "\n", 0);
  }
  bufcat(&rqst->uri, "\n", 0);
  if (len > 0) {
    bufappend(&rqst->uri, body, len);
  }

  len = rqst->uri.end - rqst->uri.data;
  p = rqst->uri.data;
  do {
    ssize_t r = send(fd, p, len, 0);
    if (r == SOCKET_ERROR) {
      if (
#if defined(_WIN32) || defined(_WIN64)
        WSAGetLastError() == WSAEWOULDBLOCK
#else
        errno == EAGAIN || errno == EWOULDBLOCK
#endif
        ) {
        poll((struct pollfd []){{ .fd = fd, .events = POLLOUT }}, 1, -1);
        continue;
      }

      return -1;
    }

    len -= r;
    p += r;
  } while(len);

  /* reset request state */
  rqst->state = http_status;
  rqst->status = 0;
  rqst->content_length = 0;
  rqst->body = NULL;
  if (rqst->buffer.start > rqst->buffer.data) {
    memmove(rqst->buffer.data, rqst->buffer.start,
            rqst->buffer.end - rqst->buffer.start + 1);
    rqst->buffer.end -= rqst->buffer.start - rqst->buffer.data;
    rqst->buffer.start = rqst->buffer.data;
  }

  return 0;
}


/* --- http_data ----------------------------------------------------------- */

static void
http_data(int fd, struct http_rqst *rqst)
{
  char *l;

  switch(rqst->state) {
    case http_status:
      l = http_readline(fd, &rqst->buffer);
      if (!l) {
        rqst->state = http_eof;
        return;
      }

      if (sscanf(l, "%*s %" SCNu16, &rqst->status) != 1)
        rqst->status = 0;
      rqst->state = http_header;
      break;

    case http_header:
      l = http_readline(fd, &rqst->buffer);
      if (!l) {
        rqst->state = http_eof;
        return;
      }

      if (*l == '\0') {
        if (!rqst->content_length) {
          rqst->body = l;
          rqst->state = http_ok;
        } else
          rqst->state = http_body;
      } else {
        int64_t s;
        if (sscanf(l, "%*[Cc]ontent-%*[Ll]ength: %" SCNi64, &s) == 1)
          rqst->content_length = s;
      }
      break;

    case http_body:
      rqst->body = http_read(fd, &rqst->buffer, rqst->content_length);
      rqst->state = http_ok;
      break;

    case http_eof:
    case http_ok:
      break;
  }
}


/* --- http_readline ------------------------------------------------------- */

static char *
http_readline(int fd, struct http_buffer *buf)
{
  char tmp[1025];
  char *p, *l;
  ssize_t n, t;
  int s;

  if (buf->end > buf->start) {
    l = buf->start;
    p = strchr(l, '\n');
    if (p) {
      *p = '\0';
      buf->start = p + 1;
      return l;
    }
  }

  t = buf->end - buf->start;
  do {
    s = poll((struct pollfd []){{ .fd = fd, .events = POLLIN }}, 1, -1);
    if (s < 0) return NULL;

    n = bufread(fd, tmp, sizeof(tmp));
    if (n <= 0) return NULL;

    p = strchr(tmp, '\n');
    if (p) {
      t += p - tmp;
    } else
      t += n;

    bufappend(buf, tmp, n);
  } while(!p);

  l = buf->start;
  buf->start += t;
  *buf->start = '\0';
  buf->start++;

  return l;
}


/* --- http_read ----------------------------------------------------------- */

static char *
http_read(int fd, struct http_buffer *buf, size_t len)
{
  char tmp[1024];
  char *l;
  ssize_t n;
  size_t got;
  int s;

  l = buf->start;
  got = buf->end - buf->start;
  if (got >= len) {
    buf->start += len;
    return l;
  }

  do {
    s = poll((struct pollfd []){{ .fd = fd, .events = POLLIN }}, 1, -1);
    if (s < 0) return NULL;

    n = bufread(fd, tmp, sizeof(tmp));
    if (n <= 0) return NULL;
    got += n;
    bufappend(buf, tmp, n);
  } while(got < len);

  l = buf->start;
  buf->start += len;
  return l;
}


/* --- bufread ------------------------------------------------------------- */

static ssize_t
bufread(int fd, char *buf, size_t count)
{
  ssize_t n;
  char *p;

  n = recv(fd, buf, count-1, 0);
  if (n == SOCKET_ERROR) return -1;
  if (n == 0) return 0;
  buf[n] = 0;

  p = buf;
  while((p = strstr(p, "\r\n"))) {
    memmove(p, p+1, n - (p - buf));
    n--;
    p++;
  }

  return n;
}


/* --- bufcat -------------------------------------------------------------- */

static int
bufcat(struct http_buffer *buf, const char *data, int quoted)
{
  static const char specials[] = "% ;?:@&=+$,<>#\"{}|\\^[]`";

  ssize_t avail;
  const char *s;
  char *d;

  s = data;
  while(*s) {
    avail = buf->len - (buf->end - buf->data);
    for(d = buf->end; *s && avail > 2; avail--) {
      if (quoted) {
        if (strchr(specials, *s)) {
          sprintf(d, "%%%02x", *s++);
          d += 3;
          avail -= 2;
        } else
          *d++ = *s++;
      } else
        *d++ = *s++;
    }

    if (avail < 3) {
      char *newbuf = (char *)realloc(buf->data, buf->len << 1);
      if (!newbuf) return ENOMEM;

      d = newbuf + (d - buf->data);
      buf->start = newbuf + (buf->start - buf->data);
      buf->data = newbuf;
      buf->len <<= 1;
    }
    *d = 0;
    buf->end = d;
  }

  return 0;
}

static int
bufappend(struct http_buffer *buf, const char *data, size_t n)
{
  ssize_t avail;

  avail = buf->len - (buf->end - buf->data);
  while (avail < n + 1) {
    char *newbuf = (char *)realloc(buf->data, buf->len << 1);
    if (!newbuf) return ENOMEM;

    buf->start = newbuf + (buf->start - buf->data);
    buf->end = newbuf + (buf->end - buf->data);
    buf->data = newbuf;
    buf->len <<= 1;
    avail = buf->len - (buf->end - buf->data);
  }

  memcpy(buf->end, data, n);
  buf->end += n;
  *buf->end = '\0';

  return 0;
}

static int
bufcat_uint64_t(struct http_buffer *buf, int neg, uint64_t i)
{
  char a[32];
  char *d;

  d = a + sizeof(a);
  *--d = 0;
  do { *--d = '0' + i%10; } while (i /= 10);
  if (neg) *--d = '-';

  return bufcat(buf, d, 0);
}
